#ifdef SYSTEM5
#include "sys/param.h"
#include "sys/config.h"
#include "sys/types.h"
#include "sys/systm.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/tty.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/buf.h"
#endif

#ifdef VERSION7
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/dir.h>
#include <sys/user.h>
#endif
#ifdef PEGASUS
#include <sys/mmu.h>
#endif

#include <sys/extypes.h>
#include <sys/socket.h>
#include <sys/exreg.h>
#include <sys/swap.h>
#include <sys/init.h>
#include <sys/exos.h>
#include <ex_ioctl.h>

extern	int ex_inited;
extern	struct exctrl ex_db;
extern	struct rmsg_area *rmsgarea;
extern	struct wmsg_area *wmsgarea;
#ifdef PEGASUS
#define vtop(x)		((long)x + (long)dscraddr(SEGKD))
#endif

#undef	DEBUG
#undef	DEBUG_NOISE
#undef	DEBUG_QUEUES

/*
 * exsetup:
 *	- perform probe level initialization
 *	- setup controller mode according to "mode"
 *	- setup message queues
 *	- send init message to exos
 *	- look at response and return value accordingly
 */
int
exsetup(mode)
	int mode;
{
	register struct init_msg *im;
	register struct msg *current, *next;
	static char magic[8] = {0xff, 0xff, 0, 0};
	char *ap;
#ifdef PEGASUS
	long addr,timeout,i;
#else
	register int i, timeout;
	long addr;
#endif

    /* link together the read ("exos to host") message queue */

	rmsgarea->ma_rlink = (long)vtop(rmsgarea->ma_rmsgs);
	current = &rmsgarea->ma_rmsgs[NET_RBUFS-1];
	rmsgarea->ma_lastr = rmsgarea->ma_rmsgs;
	for (i = 0; i < NET_RBUFS; i++) {
		next = &rmsgarea->ma_rmsgs[i];
		current->msg_hd.mh_link = (long)vtop(next);
		current->msg_hd.mh_length = sizeof(union exos_u);
		current->msg_hd.mh_status = 3;
		current->msg_link = next;
		current = next;
	}

    /* link together the write ("host to exos") message queue */

	wmsgarea->ma_wlink = (long)vtop(wmsgarea->ma_wmsgs);
	current = &wmsgarea->ma_wmsgs[NET_WBUFS-1];
	wmsgarea->ma_lastw = wmsgarea->ma_wmsgs;
	for (i = 0; i < NET_WBUFS; i++) {
		next = &wmsgarea->ma_wmsgs[i];
		current->msg_hd.mh_link = (long)vtop(next);
		current->msg_hd.mh_length = sizeof(union exos_u);
		current->msg_hd.mh_status = 0;
		current->msg_link = next;
		current = next;
	}

    /* setup init_msg data structure */

	im = (struct init_msg *)ex_db.ex_imsg;
	clear(im, sizeof(struct init_msg));	/* zap to all zero's */
	im->im_newstyle = 1;			/* use new-style init msg */
	im->im_mode = mode;			/* link level */
	im->im_hdfo[0] = im->im_hdfo[1] = 1;	/* do auto-byte/word swapping */
	im->im_junk[0] = 1;
	im->im_haddrmode = 3;			/* absolute address mode */

	/* data order test patterns */
	im->im_byteptn[0] = 1;
	im->im_byteptn[1] = 3;
	im->im_byteptn[2] = 7;
	im->im_byteptn[3] = 0xf;
	im->im_wordptn[0] = 0x103;
	im->im_wordptn[1] = 0x70f;
	im->im_longptn = 0x103070f;

	im->im_101off = im->im_101seg = 0xffff;
	im->im_nhosts = 1;
	im->im_result = im->im_nmb = im->im_nproc = im->im_nslots = 0xff;

	/* prob. here calc seg which is diff from that calc in ex_getbuf() ? */
	im->im_h2exqaddr = ((long)vtop(&wmsgarea->ma_wlink)) & 0xff0000;/*seg base*/
	im->im_h2exoff = (long)vtop(&wmsgarea->ma_wlink);
	im->im_h2extype = 0;		/* no int. when send exos a msg */
	im->im_ex2hqaddr = ((long)vtop(&rmsgarea->ma_rlink)) & 0xff0000;/*seg base*/
	im->im_ex2hoff = (long)vtop(&rmsgarea->ma_rlink);
	im->im_ex2htype = 3;
#ifdef DEBUG
	printf("exsetup: im->im_h2exoff is %x\n",(short)(im->im_h2exoff&0xffff));
	printf("exsetup: im->im_h2exqaddr is %x%x\n",(int)((im->im_h2exqaddr>>16)&0xffff),(int)(im->im_h2exqaddr&0xffff));
	printf("exsetup: im->im_ex2hoff is %x\n",(short)(im->im_ex2hoff&0xffff));
	printf("exsetup: im->im_ex2hqaddr is %x%x\n",(int)((im->im_ex2hqaddr>>16)&0xffff),(int)(im->im_ex2hqaddr&0xffff));
#endif

    /* give init_msg to exos */

#ifdef PEGASUS
	inb(PORTA);
#else
	(void) inb(PORTA);
#endif
	timeout = 1000000;
	while (((inb(PORTB) & PB_ERROR) == 0) && --timeout)
		;
	if (timeout == 0) {
		printf("ex0: board failed diagnostics\n");
		return 1;
	}

    /*
     * get physical address in exos notation:
     *	- upper 4 bits of long word are the segment number
     *	- lower 16 bits of long word are the segment offset
     */
	addr = (long)vtop(ex_db.ex_imsg);
#ifdef DEBUG
	printf("exsetup: vaddr is %x%x\n",(int)((addr>>16)&0xffff),(int)(addr&0xffff));
#endif
	for (i = 0; i < 4; i++) {
		magic[i+4] = addr;
		addr >>= 8;
	}
	for (ap = magic; ap <= &magic[7]; ap++) {
		timeout = 100000;
		while ((inb(PORTB) & PB_READY) && --timeout)
			;
		if (timeout == 0) {
			printf("ex0: board hung while sending init addr\n");
			return 1;
		}
		outb(PORTB, (*ap)&0xff);
#ifdef DEBUG
		printf("exsetup: write %x to portb\n",(*ap)&0xff);
#endif
	}

    /* wait for init to complete */

	timeout = 1000000;
	while (((im->im_result & 0xff) == 0xff) && --timeout)
		;
	if (timeout == 0) {
		printf("ex0: board hung while waiting for init to complete\n");
		return 1;
	}

    /* check result of init */

	if (im->im_result)
	{
		printf("ex0: board gave bad result 0x%x\n",im->im_result);
		return 1;
	}
#ifdef	DEBUG
	printf("[nx %o.%o, ex %o.%o] mmsize=%o ex2htype=%o\n",
		    im->im_version[0], im->im_version[1],
		    im->im_version[2], im->im_version[3],
		    im->im_mmsize,im->im_ex2htype);
#endif
	outb(PORTB, 0);
#ifdef	DEBUG
	printf("ex0: init done\n");
#endif
	ex_inited = 1;		/* tell world that board is inited */
	return 0;
}

/*
 * ex_findmsg:
 *	- first find a kludge buffer to hold result of message
 *	- find next available message on "host to exos" message queue
 *	- sleep if necessary, to get one
 */
struct msg *
ex_findmsg()
{
	register struct msg *mp;
	register struct kludge *kp;
	register int s;

	s = spl5();
	for (;;) {
		for (kp = kludge; kp < &kludge[NET_KLUDGES]; kp++) {
			if (kp->kl_state == KL_FREE) {
				kp->kl_state = KL_BUSY;
				goto findmsg;
			}
		}
		goto nightynight;
findmsg:
		mp = wmsgarea->ma_lastw;
		do {
			if (mp->msg_hd.mh_status == 0) {
				mp->msg_kludge = kp;
				mp->msg_hd.mh_status = MQ_DONE;
				mp->msg_hd.mh_reserved = 0;
				clear(&mp->msg_msg, sizeof(mp->msg_msg));
				wmsgarea->ma_lastw = mp->msg_link;
				splx(s);
#ifdef	DEBUG_QUEUES
				printf("ex_findmsg: return mp=%x%x, kp=%x\n",
				  (int)((mp>>16)&0xffff),(int)(mp&0xffff), kp);
#endif	DEBUG_QUEUES
				return mp;
			}
			mp = mp->msg_link;
		} while (mp != wmsgarea->ma_lastw);
		kp->kl_state = KL_FREE;		/* release kludge buf */
nightynight:
		ex_db.ex_state |= ST_WAITING;
		sleep((caddr_t)&ex_db, PRIBIO);
	}
}

/*
 * ex_send:
 *	- send a network message via the message queue's to the exos
 *	- set up standard header
 *	- bump board to tell it to go
 *	- wait for operation to complete and return the pointer to the
 *	  result info (kept in kludge)
 *	- must not be called at interrupt level
 */
struct kludge *
ex_send(command, mp, np, dontint)
	int command;
	register struct msg *mp;
	register struct net_msg *np;
	int dontint;
{
	register struct kludge *kp;
	register int s;

	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}

	s = spl5();
	mp->msg_hd.mh_status |= MQ_EXOS;
	np->nm_userid = (long) (kp = mp->msg_kludge);
	np->nm_request = command;
#ifdef	DEBUG_QUEUES
	printf("{ex_send: kp=%x} ", kp);
#endif	DEBUG_QUEUES

	kp->kl_state |= KL_WAITING;
	kp->kl_proc = u.u_procp;
	outb(PORTB, 0);
	if (dontint)
		sleep((caddr_t)kp, PRIBIO);	/* not interruptable */
	else
		sleep((caddr_t)kp, PZERO+1);	/* interruptable */
	splx(s);

#ifdef	DEBUG_QUEUES
	printf("{ex_send: reply=%x} ", kp->kl_reply);
#endif	DEBUG_QUEUES
#ifdef	DEBUG_NOISE
	if (kp->kl_reply)
		printf("ex0: reply = %x\n", kp->kl_reply);
#endif	DEBUG_NOISE
	return kp;
}
